from types import TracebackType
from typing import Optional, Protocol, Type

from lonelypsp.auth.config import BadAuthResult
from lonelypsp.tracing.shared.handle_trusted_notify import HandledTrustedNotify
from lonelypsp.tracing.shared.tracing_and_followup import TracingAndFollowup


class StatelessTracingNotifyStart(Protocol):
    """Entry context manager for the subscriber side of a NOTIFY"""

    def __enter__(self) -> "StatelessTracingNotifyStartFirst":
        """Called to initialize the trace on the subscriber side for a NOTIFY"""

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """
        Called when the subscriber side of this trace is completely done, which
        may be due to cancellation, on a best-effort basis (i.e., as good as
        `with` could do but not better)
        """


class StatelessTracingNotifyStartFirst(Protocol):
    """First object run on the subscriber (and overall)"""

    def on_start_without_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> "StatelessTracingNotifyOnHashed":
        """Called immediately after the `lonelypsc` library is told to make
        a NOTIFY call, given that the client did not provide the sha512 hash (so
        lonelypsc needs to compute it)

        Args:
            topic (bytes): the topic to write to
            length (int): the length of the data to write
            filelike (bool): whether the data is provided as a file-like object (True)
                or as a bytes object (False). When given a bytes object all operations
                are performed in memory regardless of the length

        Returns:
            StatelessTracingNotifyOnHashed: the next object to use
        """

    def on_start_with_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> "StatelessTracingNotifyOnSending":
        """Called immediately after the `lonelypsc` library is told to make
        a NOTIFY call, given that the client provided the sha512 hash (so
        lonelypsc does not need to compute it)

        Args:
            topic (bytes): the topic to write to
            length (int): the length of the data to write
            filelike (bool): whether the data is provided as a file-like object (True)
                or as a bytes object (False). When given a bytes object all operations
                are performed in memory regardless of the length

        Returns:
            StatelessTracingNotifyOnSending: the next object to use
        """


class StatelessTracingNotifyOnHashed(Protocol):
    """Run on the subscriber, omitted if the hash was provided to lonelypsc"""

    def on_hashed(self) -> "StatelessTracingNotifyOnSending":
        """Called after the lonelypsc library has finished computing the hash
        of the message and is moving to getting ready to send the message to
        the server
        """


class StatelessTracingNotifyOnSending(Protocol):
    """Run on the subscriber immediately prior to transferring to the broadcaster"""

    def on_sending_request(
        self, /, *, broadcaster: str, identifier: bytes
    ) -> TracingAndFollowup["StatelessTracingNotifyOnResponseReceived"]:
        """Called immediately prior to the subscriber sending the request to
        the broadcaster. Must return the tracing data to send to the broadcaster
        to pick up

        Args:
            broadcaster (str): the broadcaster to send the request to
            identifier (bytes): the arbitrary identifier the subscriber assigned
                to this request; this will be passed to the broadcaster without
                having to store it in the tracing data, and can be assumed to
                be globally unique

        Returns:
            tracing: the tracing data to send, max 2^16-1 bytes, may be empty
            followup: the next object on the subscriber side
        """


class StatelessTracingNotifyOnReceived(Protocol):
    """Entry context manager for the broadcaster side of a NOTIFY"""

    def __enter__(self) -> "StatelessTracingNotifyOnReceivedFirst":
        """Called to initialize the trace on the broadcaster side for a NOTIFY"""

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Called when the broadcaster side of this trace is done, which may be due
        to cancellation, on a best-effort basis (i.e., as good as `with` could
        do but not better)
        """


class StatelessTracingNotifyOnReceivedFirst(Protocol):
    """First object run on the broadcaster (but not overall)"""

    def on_received(self) -> "StatelessTracingNotifyOnAuthResult":
        """Called immedaitely upon detecting a NOTIFY over the stateless
        protocol, before the body has been parsed or authorization has been
        checked. Generally only useful for collecting timing data if the
        request is accepted
        """


class StatelessTracingNotifyOnAuthResult(Protocol):
    """Run on the broadcaster"""

    def on_bad_request(self) -> None:
        """Called if the request is malformed or otherwise invalid, terminating
        the chain without a useful response to the subscriber
        """

    def on_bad_auth_result(self, /, *, result: BadAuthResult) -> None:
        """Called if the authorization check failed, terminating the chain
        without a useful response to the subscriber
        """

    def on_auth_tentatively_accepted(self) -> "StatelessTracingNotifyOnAuthVerified":
        """Called if the authorization check passed, meaning that if the hashes
        all actually match what was sent, then the request was sent by the subscriber
        """


class StatelessTracingNotifyOnAuthVerified(Protocol):
    """Run on the broadcaster"""

    def on_auth_mismatch(self) -> None:
        """Called if the authorization provided doesn't actually match what was
        sent. This will terminate the chain without a useful response to the
        subscriber
        """

    def on_auth_accepted(
        self, topic: bytes, length: int, identifier: bytes, tracing: bytes
    ) -> "HandledTrustedNotify[StatelessTracingNotifyOnSendingResponse]":
        """Called if the authorization header matches what was sent, so
        the broadcaster is moving onto actually processing the notification.

        Args:
            topic (bytes): the topic to find subscribers of
            length (int): the length of the message to send
            identifier (bytes): the arbitrary identifier the subscriber assigned
                to this request, can be assumed to be globally unique if authorized
                subscribers can be trusted, otherwise a new identifier needs to be
                generated if uniqueness is required. In other words, this is globally
                unique iff only trusted internal services can pass authorization
            tracing (bytes): the tracing data sent by the subscriber
        """


class StatelessTracingNotifyOnSendingResponse(Protocol):
    """Runs on the broadcaster"""

    def on_sending_response(self) -> bytes:
        """Called as close as possible to actually sending the `RESPONSE_NOTIFY`
        message to the subscriber. This is necessarily before producing the
        authorization header, which will sign the tracing data, and actually
        formatting the response body, which will include the tracing data.

        Returns:
            bytes: the tracing data to send to the subscriber
        """


class StatelessTracingNotifyOnResponseReceived(Protocol):
    """Runs on the subscriber after receiving the response from the broadcaster
    but before reading the body or verifying the authorization
    """

    def on_network_error(self) -> "StatelessTracingNotifyOnRetryDetermined":
        """Called if the subscriber did not receive a response from the broadcaster
        because either subscriber could not reach the broadcaster, the broadcaster
        did not understand the request or experienced an error that prevented it
        from forming a sensible response, or the broadcasters response could not
        reach the subscriber
        """
        ...

    def on_response_received(
        self, /, *, status_code: int
    ) -> "StatelessTracingNotifyOnResponseAuthResult":
        """Called when the subscriber has received what should be a `RESPONSE_NOTIFY`
        from the broadcaster, but before the body has been read or the authorization
        has been checked

        Args:
            status_code (int): the HTTP status code of the response, which isn't
                used for anything but logging
        """


class StatelessTracingNotifyOnRetryDetermined(Protocol):
    def on_retry_prevented(self) -> None:
        """Called if the subscriber does not want to attempt to retry this
        type of error
        """

    def on_retries_exhausted(self) -> None:
        """Called if the subscriber wants to retry this request but has exhausted
        its retry limit
        """

    def on_waiting_to_retry(self) -> "StatelessTracingNotifyOnRetryDetermined":
        """Called if the subscriber will wait a bit before trying the next
        broadcaster, immediately before sleeping
        """

    def on_retrying(self) -> StatelessTracingNotifyOnSending:
        """Called if the subscriber will try to notify this or a different
        broadcaster with the same message. A new authorization header will
        be created and there is an opportunity to change the tracing data.
        """


class StatelessTracingNotifyOnResponseAuthResult(Protocol):
    """Runs on the subscriber"""

    def on_bad_response(self) -> StatelessTracingNotifyOnRetryDetermined:
        """Called if the response was malformed or otherwise invalid, so the
        subscriber didn't get any useful tracing data. Will determine if
        retrying is appropriate and provide that info to the next object.
        """

    def on_bad_auth_result(
        self, /, *, result: BadAuthResult
    ) -> StatelessTracingNotifyOnRetryDetermined:
        """Called if the response authorization failed, meaning it may not have
        come from the broadcaster and there is no useful tracing data.
        Will determine if retrying is appropriate and provide that info to the
        next object.
        """

    def on_response_notify_accepted(
        self, /, *, tracing: bytes, num_subscribers: int
    ) -> None:
        """Called if the response was correctly formed and passed authorization,
        meaning the subscriber has received the final tracing data from the
        broadcaster and is about to return from the lonelypsc library

        Args:
            tracing (bytes): the tracing data sent by the broadcaster
            num_subscribers (int): the number of subscribers that received the
                notification, which is often useful for the caller when they know
                how many they were expecting
        """



from enum import IntEnum, auto
from typing import (
    TYPE_CHECKING,
    Literal,
    Optional,
    Protocol,
    Type,
)

from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag


class AuthResult(IntEnum):
    """Distinguishes the different possible results of an authorization check"""

    OK = auto()
    """the request is allowed"""

    UNAUTHORIZED = auto()
    """the authorization header is required but not provided"""

    FORBIDDEN = auto()
    """the authorization header is provided but invalid"""

    UNAVAILABLE = auto()
    """a service is required to check this isn't available"""


BadAuthResult = Literal[
    AuthResult.UNAUTHORIZED, AuthResult.FORBIDDEN, AuthResult.UNAVAILABLE
]


class ToBroadcasterAuthConfig(Protocol):
    """Handles verifying requests from a subscriber to this broadcaster or
    producing the authorization header when contacting other broadcasters
    """

    async def setup_to_broadcaster_auth(self) -> None:
        """Prepares this authorization instance for use. If the
        to broadcaster auth config is not re-entrant (i.e., it cannot
        be used by two clients simultaneously), it must detect this and error
        out.
        """

    async def teardown_to_broadcaster_auth(self) -> None:
        """Cleans up this authorization instance after use. This is called when a
        client is done using the auth config, and should release any resources
        it acquired during `setup_to_broadcaster_auth`.
        """

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to subscribe
        the given url to the given topic.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            exact (bytes): the exact topic they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to
        (un)subscribe a specific url to a specific topic

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            url (str): the url that will receive notifications
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any. Always None for unsubscribes
            exact (bytes): the exact topic they want to receive messages from
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to subscribe
        the given url to any message sent to a topic which matches the glob.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            glob (str): the glob pattern they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to
        (un)subscribe a specific url to a specific glob of topics

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            url (str): the url that will receive notifications
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            glob (str): a glob for the topics that they want to receive notifications from
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to fanout
        a notification on a specific topic. As the message may be very large, only
        the sha512 of the message is used for authorization.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            topic (bytes): the topic that the message is being sent to
            identifier (bytes): an arbitrary identifier for this message assigned
                by the subscriber, max 255 bytes
            message_sha512 (bytes): the sha512 of the message being sent
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to fanout a
        notification on a specific topic.

        As we support very large messages, for authorization only the SHA-512 of
        the message should be used, which will be fully verified before any
        notifications go out.

        Note that in websockets where compression is enabled, the sha512 is
        of the compressed content, as we cannot safely decompress the data (and
        thus compute the decompressed sha512) unless we know it is safe, at which
        point a second check would be redundant.

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            topic (bytes): the topic that the message is being sent to
            identifier (bytes): an arbitrary identifier for this message assigned
                by the subscriber, max 255 bytes
            message_sha512 (bytes): the sha512 of the message being sent
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to configure
        a stateful connection. This is the first packet in the stateful protocol

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            subscriber_nonce (bytes): the 32 random bytes the subscriber is
                contributing toward the connection nonce
            enable_zstd (bool): whether to enable zstd compression
            enable_training (bool): whether to enable training mode
            initial_dict (int): the initial dictionary to use

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to configure
        a stateful connection with a subscriber.

        Args:
            message (S2B_Configure): the configure message they sent
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        """Produces the authorization header sent to the broadcaster to check
        the subscriptions for a specific url.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            url (str): the url the subscriber is checking
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to check
        the subscriptions for a specific url.

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            url (str): the url whose subscriptions are being checked
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        """Produces the authorization header sent to the broadcaster to replace
        the subscriptions for a specific url.

        Unlike with the checking side which might compare the user being
        authenticated with vs the topics, there is generally no reason to need
        to view the specific globs/topics that are being subscribed to for
        generating the authorization token, as if they are not valid it will
        be caught by the broadcaster

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            url (str): the url the subscriber is setting
            strong_etag (StrongEtag): the strong etag of the subscriptions being set
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to replace
        the subscriptions for a specific url.

        Ideally the authorization would not need to actually iterate the topics
        and globs, but in practice that is too great a restriction, so instead
        the iterable is async, single-use, and can detect if it was unused, allowing
        the implementation the maximum flexibility to make performance optimizations
        while still allowing the obvious desired case of some users can only subscribe
        to certain prefixes

        WARN: when this function returns, `subscriptions` will no longer be usable

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            url (str): the url whose subscriptions are being set
            strong_etag (StrongEtag): the strong etag that will be verified before
                actually setting subscriptions, but may not have been verified yet.
            subscriptions (SetSubscriptionsInfo): the subscriptions to set
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to acknowledge
        that the subscriber has finished receiving the given part id of the given arbitrarily
        identified message and is ready to receive more. This is used on stateful connections
        which have simultaneous, continuous receive and send capabilities, and is used as a
        form of backpressure. Note that the broadcaster might send e.g. 3 messages before
        waiting for the ack on the first one, to keep the connection reasonably saturated

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the broadcaster on the first `RECEIVE_STREAM`
            part_id (int): the part id of the message received
            url (str): websocket:<nonce>:<ctr> which helps detect desyncs
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to acknowledge
        that the subscriber has finished receiving the given part id of the given arbitrarily
        identified message and is ready to receive more.

        Args:
            url (str): websocket:<nonce>:<ctr> which helps detect desyncs
            message (S2B_ContinueReceive): the continue receive message from the subscriber
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to confirm
        that the subscriber has received the entire message with the given identifier.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the broadcaster
            num_subscribers (int): the number of subscribers that received the message
            url (str): the url the subscriber received the message on; for websockets, this
                is a unique identifier to this request
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to confirm
        that the subscriber has received the entire message with the given identifier.

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the broadcaster
            num_subscribers (int): the number of subscribers that received the message
            url (str): the url the subscriber received the message on; for websockets, this
                is a unique identifier to this request
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        """Produces the authorization header to send to the broadcaster to confirm
        that the subscriber received that it may have missed a message on the given
        topic via the given url at approximately the given time.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            topic (bytes): the topic that one or more messages may have been missed on
            url (str): the url the subscriber missed the message on
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the broadcaster to confirm
        that the subscriber received that it may have missed a message on the given
        topic via the given url at approximately the given time.

        Args:
            tracing (bytes): the tracing data from the subscriber, may be empty
            topic (bytes): the topic that one or more messages may have been missed on
            url (str): the url the subscriber missed the message on
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """


class ToSubscriberAuthConfig(Protocol):
    """Handles verifying requests from a broadcaster to this subscriber or
    producing the authorization header when contacting subscribers
    """

    async def setup_to_subscriber_auth(self) -> None:
        """Prepares this authorization instance for use. If the to subscriber auth
        config is not re-entrant (i.e., it cannot be used by two clients
        simultaneously), it must detect this and error out.
        """

    async def teardown_to_subscriber_auth(self) -> None:
        """Cleans up this authorization instance after use. This is called when a
        client is done using the auth config, and should release any resources it
        acquired during `setup_to_subscriber_auth`.
        """

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber a message
        with the given sha512 on the given topic at approximately the given
        time.

        When using websockets, the url is of the form "websocket:<nonce>:<ctr>",
        where more details are described in the websocket endpoints
        documentation. What's important is that the recipient can either verify
        the url is what they expect or the url is structured such that it is
        unique if _either_ party is acting correctly, meaning replay attacks are
        limited to a single target (i.e., we structurally disallow replaying a
        message sent from Bob to Alice via pretending to be Bob to Charlie, as
        Charlie will be able to tell that message was intended for not-Charlie).

        Note that the reverse is not promised (i.e., broadcasters do not know which
        broadcaster the subscriber meant to contact), but assuming the number of
        broadcasters is much smaller than the number of subscribers, this is less
        of an issue to coordinate.

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            url (str): the url that will receive the notification
            topic (bytes): the topic that the message is being sent to
            message_sha512 (bytes): the sha512 of the message being sent
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to a subscriber to receive a message
        from a broadcaster on a topic.

        As we support very large messages, for authorization only the SHA-512 of
        the message should be used, which will be fully verified.

        Broadcasters act as subscribers for receiving messages when a subscriber
        is connected via websocket, so it can forward messages sent to other
        broadcasters.

        Args:
            tracing (bytes): the tracing data the broadcaster provided, may be empty
            topic (bytes): the topic the message claims to be on
            message_sha512 (bytes): the sha512 of the message being received
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to indicate
        that it may have missed a message on the given topic. The message is being sent at
        approximately the given time, which is unrelated to when the message they
        missed was sent.

        The contents of the message are not sent nor necessarily available; this
        is just to inform the subscriber that they may have missed a message.
        They may have their own log that they can recovery the message with if
        necessary.

        When sending this over a websocket, the recovery url is of the form
        `websocket:<nonce>:<ctr>`, where more details can be found in the
        stateful documentation in lonelypsp

        Args:
            tracing (bytes): the tracing data to send to the broadcaster, may be empty
            recovery (str): the url that will receive the missed message
            topic (bytes): the topic that the message was on
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to a subscriber via the given
        recovery url to indicate it may have missed a message on the given topic

        Broadcasters act as subscribers for receiving messages when a subscriber
        is connected via websocket, so it can forward messages sent to other
        broadcasters.

        Args:
            tracing (bytes): the tracing data the broadcaster provided, may be empty
            recovery (str): the url the missed message was sent to
            topic (bytes): the topic the message was on
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to confirm
        the exact subscription they sent was accepted.

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            exact (bytes): the exact topic they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to confirm
        the exact subscription they sent was accepted.

        Args:
            tracing (bytes): the tracing data from the broadcaster, may be empty
            url (str): the url that will receive notifications
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            exact (bytes): the exact topic they want to receive messages from
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to confirm
        the glob subscription they sent was accepted.

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            glob (str): the glob pattern they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to confirm
        the glob subscription they sent was accepted.

        Args:
            tracing (bytes): the tracing data from the broadcaster, may be empty
            url (str): the url that will receive notifications
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            glob (str): a glob for the topics that they want to receive notifications from
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns: AuthResult
        """

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to confirm
        the notification they sent was accepted.

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the subscriber
            subscribers (int): the number of subscribers that received the message
            topic (bytes): the topic that the message is being sent to
            message_sha512 (bytes): the sha512 of the message being sent
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to confirm
        the notification they sent was accepted.

        Args:
            tracing (bytes): the tracing data from the broadcaster, may be empty
            identifier (bytes): the arbitrary identifier of the message being received.
                this identifier was assigned by the subscriber
            subscribers (int): the number of subscribers that received the message
            topic (bytes): the topic that the message is being sent to
            message_sha512 (bytes): the sha512 of the message being sent
            authorization (str, None): the authorization header they provided
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to confirm
        the stateful configure message they sent was accepted.

        Args:
            broadcaster_nonce (bytes): the nonce that the broadcaster will send in the
                confirm configure message
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to a subscriber to confirm the
        stateful configure message they sent was accepted.

        Args:
            message (B2S_ConfirmConfigure): the confirm configure message from the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        authorization: Optional[str],
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to enable
        zstd compression with the given parameters over a stateful connection.

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            url (str): the url the subscriber is subscribing to
            compressor_identifier (int): the identifier of the zstd compressor
            compression_level (int): the compression level to use
            min_size (int): the minimum size of messages to compress
            max_size (int): the maximum size of messages to compress, up to 2^64 - 1
            authorization (str, None): the authorization header they provided
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to enable
        zstd compression with the given parameters over a stateful connection.

        Args:
            url (str): websocket:<nonce>:<ctr>, unique to this request
            message (B2S_EnableZstdPreset): the enable zstd preset message from the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to enable
        zstd compression with the given parameters over a stateful connection.

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            url (str): the url the subscriber is subscribing to
            compressor_identifier (bytes): the identifier of the zstd compressor
            compression_level (int): the compression level to use
            min_size (int): the minimum size of messages to compress
            max_size (int): the maximum size of messages to compress, up to 2^64 - 1
            sha512 (bytes): the sha512 of the compressor dictionary, which is used in
                place of the actual dictionary data for authorization (as the dictionary
                could be pretty large)
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to enable
        zstd compression with the given parameters over a stateful connection.

        Args:
            url (str): websocket:<nonce>:<ctr>, unique to this request
            message (B2S_EnableZstdCustom): the enable zstd custom message from the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        """Produces the authorization header to send to the subscriber to disable
        zstd compression previously enabled with the given identifier

        Args:
            tracing (bytes): the tracing data to send to the subscriber, may be empty
            compressor_identifier (int): the identifier of the zstd compressor
            url (str): the url the subscriber is subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        """Checks the authorization header posted to the subscriber to disable
        zstd compression previously enabled with the given identifier

        Args:
            url (str): websocket:<nonce>:<ctr>, unique to this request
            message (B2S_DisableZstdCustom): the disable zstd custom message from the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns: AuthResult
        """


class AuthConfig(ToBroadcasterAuthConfig, ToSubscriberAuthConfig, Protocol): ...


class AuthConfigFromParts:
    """Convenience class to combine an incoming and outgoing auth config into an
    auth config
    """

    def __init__(
        self,
        to_broadcaster: ToBroadcasterAuthConfig,
        to_subscriber: ToSubscriberAuthConfig,
    ):
        self.to_broadcaster = to_broadcaster
        self.to_subscriber = to_subscriber

    async def setup_to_broadcaster_auth(self) -> None:
        await self.to_broadcaster.setup_to_broadcaster_auth()

    async def teardown_to_broadcaster_auth(self) -> None:
        await self.to_broadcaster.teardown_to_broadcaster_auth()

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_notify(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_notify_allowed(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_stateful_configure(
            tracing=tracing,
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
        )

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        return await self.to_broadcaster.is_stateful_configure_allowed(
            message=message, now=now
        )

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_check_subscriptions(
            tracing=tracing, url=url, now=now
        )

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        return await self.to_broadcaster.is_check_subscriptions_allowed(
            tracing=tracing, url=url, now=now, authorization=authorization
        )

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_set_subscriptions(
            tracing=tracing, url=url, strong_etag=strong_etag, now=now
        )

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_set_subscriptions_allowed(
            tracing=tracing,
            url=url,
            strong_etag=strong_etag,
            subscriptions=subscriptions,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_stateful_continue_receive(
            tracing=tracing, identifier=identifier, part_id=part_id, url=url, now=now
        )

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        return await self.to_broadcaster.is_stateful_continue_receive_allowed(
            url=url, message=message, now=now
        )

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_confirm_receive(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
        )

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_confirm_receive_allowed(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.to_broadcaster.authorize_confirm_missed(
            tracing=tracing, topic=topic, url=url, now=now
        )

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_broadcaster.is_confirm_missed_allowed(
            tracing=tracing,
            topic=topic,
            url=url,
            now=now,
            authorization=authorization,
        )

    async def setup_to_subscriber_auth(self) -> None:
        await self.to_subscriber.setup_to_subscriber_auth()

    async def teardown_to_subscriber_auth(self) -> None:
        await self.to_subscriber.teardown_to_subscriber_auth()

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_receive(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
        )

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_subscriber.is_receive_allowed(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
            authorization=authorization,
        )

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_missed(
            tracing=tracing, recovery=recovery, topic=topic, now=now
        )

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_subscriber.is_missed_allowed(
            tracing=tracing,
            recovery=recovery,
            topic=topic,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_confirm_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_subscriber.is_confirm_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_confirm_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.to_subscriber.is_confirm_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_confirm_notify(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.to_subscriber.is_confirm_notify_allowed(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            authorization=authorization,
            now=now,
        )

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_stateful_confirm_configure(
            broadcaster_nonce=broadcaster_nonce, tracing=tracing, now=now
        )

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        return await self.to_subscriber.is_stateful_confirm_configure_allowed(
            message=message, now=now
        )

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        authorization: Optional[str],
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_stateful_enable_zstd_preset(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            authorization=authorization,
            now=now,
        )

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        return await self.to_subscriber.is_stateful_enable_zstd_preset_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_stateful_enable_zstd_custom(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            sha512=sha512,
            now=now,
        )

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        return await self.to_subscriber.is_stateful_enable_zstd_custom_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        return await self.to_subscriber.authorize_stateful_disable_zstd_custom(
            tracing=tracing,
            compressor_identifier=compressor_identifier,
            url=url,
            now=now,
        )

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        return await self.to_subscriber.is_stateful_disable_zstd_custom_allowed(
            url=url, message=message, now=now
        )


if TYPE_CHECKING:
    _: Type[AuthConfig] = AuthConfigFromParts



from typing import (
    TYPE_CHECKING,
    Collection,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    Union,
    cast,
)

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import (
    GeneratorContextManager,
    parse_expanded_headers,
    parse_minimal_message_headers,
)
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_ReceiveStreamStartUncompressed:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields

    This type is for when x-part-id is 0 and x-compressor is 0
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    identifier: bytes
    """an arbitrary identifier for this notification assigned by the broadcaster; max 64 bytes
    """

    part_id: Literal[None]
    """discriminator value. We reinterpret a part id of 0 to None as we cannot
    type "strictly positive integers" in python for the counterpart to `Literal[0]`
    """

    topic: bytes
    """the topic of the message"""

    compressor_id: Literal[None]
    """discriminator value. We reinterpret a compressor id of 0 to None as we cannot
    type "strictly positive integers" in python for the counterpart to `Literal[0]`
    """

    uncompressed_length: int
    """the number of bytes that comprise the notification body"""

    unverified_uncompressed_sha512: bytes
    """The unverified sha512 hash of the entire uncompressed notification body, 64 bytes"""

    payload: bytes
    """The first part of the notification in uncompressed form"""


@fast_dataclass
class B2S_ReceiveStreamStartCompressed:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields

    This type is for when x-part-id is 0 and x-compressor is not 0
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    identifier: bytes
    """an arbitrary identifier for this notification assigned by the broadcaster; max 64 bytes
    """

    part_id: Literal[None]
    """discriminator value. We reinterpret a part id of 0 to None as we cannot
    type "strictly positive integers" in python for the counterpart to `Literal[0]`
    """

    topic: bytes
    """the topic of the message"""

    compressor_id: int
    """a positive value indicating which compressor was used to compress the message"""

    compressed_length: int
    """the number of bytes that comprise the compressed notification body"""

    decompressed_length: int
    """when decompressing the compressed data, the number of bytes that should be produced"""

    unverified_compressed_sha512: bytes
    """the unverified sha512 hash of the entire compressed notification body, 64 bytes"""

    payload: bytes
    """the first part of the notification in compressed form; the compression is over the
    entire notification body, so this is probably not decompressible by itself
    """


@fast_dataclass
class B2S_ReceiveStreamContinuation:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields

    This type is for when x-part-id is not 0
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM]
    """discriminator value"""

    authorization: Optional[str]
    """url: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    identifier: bytes
    """an arbitrary identifier for this notification assigned by the broadcaster; max 64 bytes
    """

    part_id: int
    """a positive value indicating the part number of the message. never sent out of order"""

    payload: bytes
    """the additional payload data for the notification"""


B2S_ReceiveStream = Union[
    B2S_ReceiveStreamStartUncompressed,
    B2S_ReceiveStreamStartCompressed,
    B2S_ReceiveStreamContinuation,
]


_basic_headers: Collection[str] = (
    "authorization",
    "x-tracing",
    "x-identifier",
    "x-part-id",
)

_start_headers: Collection[str] = (
    "x-topic",
    "x-compressor",
    "x-compressed-length",
    "x-decompressed-length",
    "x-compressed-sha512",
)

_total_start_headers: Collection[str] = cast(List[str], _basic_headers) + cast(
    List[str], _start_headers
)


class B2S_ReceiveStreamParser:
    """Satisfies B2S_MessageParser[B2S_ReceiveStream]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_ReceiveStream:
        assert type == BroadcasterToSubscriberStatefulMessageType.RECEIVE_STREAM
        headers: Dict[str, bytes] = dict()

        if (flags & PubSubStatefulMessageFlags.MINIMAL_HEADERS) != 0:
            with GeneratorContextManager(
                parse_minimal_message_headers(payload)
            ) as parser:
                headers["authorization"] = next(parser)
                headers["x-tracing"] = next(parser)
                headers["x-identifier"] = next(parser)

                part_id_bytes = next(parser)
                if len(part_id_bytes) > 8:
                    raise ValueError("x-part-id must be at most 8 bytes")

                headers["x-part-id"] = part_id_bytes

                part_id = int.from_bytes(part_id_bytes, "big")

                if part_id == 0:
                    for header_name in _start_headers:
                        headers[header_name] = next(parser)
        else:
            headers = parse_expanded_headers(payload)

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        identifier = headers["x-identifier"]
        if len(identifier) > 64:
            raise ValueError("x-identifier must be at most 64 bytes")

        part_id_bytes = headers["x-part-id"]
        if len(part_id_bytes) > 8:
            raise ValueError("x-part-id must be at most 8 bytes")

        part_id = int.from_bytes(part_id_bytes, "big")

        if part_id > 0:
            return B2S_ReceiveStreamContinuation(
                type=type,
                authorization=authorization,
                tracing=tracing,
                identifier=identifier,
                part_id=part_id,
                payload=payload.read(-1),
            )

        topic = headers["x-topic"]

        compressor_id_bytes = headers["x-compressor"]
        if len(compressor_id_bytes) > 8:
            raise ValueError("x-compressor must be at most 8 bytes")

        compressor_id = int.from_bytes(compressor_id_bytes, "big")
        compressed_length_bytes = headers["x-compressed-length"]
        if len(compressed_length_bytes) > 8:
            raise ValueError("x-compressed-length must be at most 8 bytes")

        compressed_length = int.from_bytes(compressed_length_bytes, "big")

        decompressed_length_bytes = headers["x-decompressed-length"]
        if len(decompressed_length_bytes) > 8:
            raise ValueError("x-decompressed-length must be at most 8 bytes")

        decompressed_length = int.from_bytes(decompressed_length_bytes, "big")

        compressed_sha512 = headers["x-compressed-sha512"]
        if len(compressed_sha512) != 64:
            raise ValueError("x-compressed-sha512 must be 64 bytes")

        if compressor_id == 0:
            if compressed_length != decompressed_length:
                raise ValueError(
                    "decompressed length must equal compressed length when x-compressor is 0"
                )

            return B2S_ReceiveStreamStartUncompressed(
                type=type,
                authorization=authorization,
                tracing=tracing,
                identifier=identifier,
                part_id=None,
                topic=topic,
                compressor_id=None,
                uncompressed_length=compressed_length,
                unverified_uncompressed_sha512=compressed_sha512,
                payload=payload.read(-1),
            )

        return B2S_ReceiveStreamStartCompressed(
            type=type,
            authorization=authorization,
            tracing=tracing,
            identifier=identifier,
            part_id=None,
            topic=topic,
            compressor_id=compressor_id,
            compressed_length=compressed_length,
            decompressed_length=decompressed_length,
            unverified_compressed_sha512=compressed_sha512,
            payload=payload.read(-1),
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_ReceiveStream]] = B2S_ReceiveStreamParser


def serialize_b2s_receive_stream(
    msg: B2S_ReceiveStream, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ReceiveStream]"""
    authorization_bytes = (
        msg.authorization.encode("utf-8") if msg.authorization is not None else b""
    )

    if msg.part_id is not None:
        return serialize_simple_message(
            type=msg.type,
            header_names=_basic_headers,
            header_values=(
                authorization_bytes,
                msg.tracing,
                msg.identifier,
                int_to_minimal_unsigned(msg.part_id),
            ),
            payload=msg.payload,
            minimal_headers=minimal_headers,
        )

    if msg.compressor_id is None:
        return serialize_simple_message(
            type=msg.type,
            header_names=_total_start_headers,
            header_values=(
                authorization_bytes,
                msg.tracing,
                msg.identifier,
                b"\x00",
                msg.topic,
                b"\x00",
                int_to_minimal_unsigned(msg.uncompressed_length),
                int_to_minimal_unsigned(msg.uncompressed_length),
                msg.unverified_uncompressed_sha512,
            ),
            payload=msg.payload,
            minimal_headers=minimal_headers,
        )

    return serialize_simple_message(
        type=msg.type,
        header_names=_total_start_headers,
        header_values=(
            authorization_bytes,
            msg.tracing,
            msg.identifier,
            b"\x00",
            msg.topic,
            int_to_minimal_unsigned(msg.compressor_id),
            int_to_minimal_unsigned(msg.compressed_length),
            int_to_minimal_unsigned(msg.decompressed_length),
            msg.unverified_compressed_sha512,
        ),
        payload=msg.payload,
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_ReceiveStream] = serialize_b2s_receive_stream



from typing import Generic, Iterable, List, Optional, TypeVar


class BoundedDequeFullError(Exception):
    """The exception raised if a bounded deque is full and an element is
    attempted to be added"""

    pass


T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")


class BoundedDequeInPlaceIter(Generic[T_co]):
    def __init__(self, parent: "BoundedDeque[T_co]") -> None:
        self.parent = parent
        self.offset = 0

    def __iter__(self) -> "BoundedDequeInPlaceIter[T_co]":
        return self

    def __next__(self) -> T_co:
        if self.offset >= self.parent.length:
            raise StopIteration
        idx = (self.parent.start + self.offset) % len(self.parent.data)
        self.offset += 1
        result = self.parent.data[idx]
        assert result is not None, "invariant violated"
        return result


class BoundedDeque(Generic[T]):
    """Functions similarly to collections.deque, but raises an error if full instead
    of silently dropping elements, plus uses the expected implementation
    (compact array) rather than the double linked list from cpython. Without
    benchmarking this is probably better on PyPy but worse in CPython

    The implementation is guarranteed in this api, meaning you can e.g.
    use resize safely

    Not thread-safe, not process-safe

    append: amortized O(1)
    pop: O(1)
    appendleft: amortized O(1)
    popleft: O(1)
    iteration: O(n), at most one out of order access assuming no resizes
    """

    def __init__(
        self, /, iterable: Optional[Iterable[T]] = None, *, maxlen: Optional[int] = None
    ) -> None:
        self.data: List[Optional[T]] = list(iterable) if iterable is not None else []
        """the data in the deque which will be resized exponentially until
        it reaches maxlen (if any)
        """

        self.start = 0
        """the first non-None entry in data"""

        self.length = len(self.data)
        """the number of non-None entries in data, all after start, wrapping
        if necessary
        """

        self.max_size = maxlen
        """the maximum allowed size, or None if unbounded"""

        if self.max_size is not None and len(self.data) > self.max_size:
            raise ValueError("iterable larger than maxlen")

    def __len__(self) -> int:
        return self.length

    def __bool__(self) -> bool:
        return bool(self.length)

    def __getitem__(self, idx: int) -> T:
        if idx < 0 or idx >= self.length:
            raise IndexError("index out of range")
        result = self.data[(self.start + idx) % len(self.data)]
        assert result is not None, "invariant violated"
        return result

    def __setitem__(self, idx: int, val: T) -> None:
        if idx < 0 or idx >= self.length:
            raise IndexError("index out of range")
        self.data[(self.start + idx) % len(self.data)] = val

    def __repr__(self) -> str:
        return f"BoundedDeque({list(self)!r}, maxlen={self.max_size!r})"

    def __str__(self) -> str:
        return f"BoundedDeque{list(self)}; maxlen={self.max_size!r}, start={self.start}, length={len(self)}, capacity={len(self.data)}"

    def resize(self, min_size: int) -> None:
        """Attempts to resize the underlying array to have at least the given
        size. Raises BoundedDequeFullError if min_size > self.max_size
        """
        if len(self.data) >= min_size:
            return

        if self.max_size is not None and min_size > self.max_size:
            raise BoundedDequeFullError

        new_data: List[Optional[T]] = [None] * min_size
        idx = self.start
        for new_idx in range(self.length):
            new_data[new_idx] = self.data[idx]
            idx += 1
            if idx == len(self.data):
                idx = 0
        self.data = new_data
        self.start = 0

    def ensure_space_for(self, n: int) -> None:
        """Ensures that there is space for n more elements in the deque.
        Raises BoundedDequeFullError if this would require expanding beyond
        min_size.

        Prefers to expand the array exponentially, but respects max_size
        """
        assert n >= 0
        min_size = self.length + n
        if min_size <= len(self.data):
            return

        next_pow_2 = 1 << (min_size - 1).bit_length()
        if self.max_size is None or next_pow_2 <= self.max_size:
            self.resize(next_pow_2)
            return

        if min_size > self.max_size:
            raise BoundedDequeFullError

        self.resize(self.max_size)

    def append(self, val: T) -> None:
        """Appends an element to the end of the deque. Raises
        BoundedDequeFullError if the deque is full
        """
        self.ensure_space_for(1)
        idx = (self.start + self.length) % len(self.data)
        self.data[idx] = val
        self.length += 1

    def pop(self) -> T:
        """Pops the last element from the deque. Raises IndexError if the
        deque is empty
        """
        if self.length == 0:
            raise IndexError("pop from an empty deque")
        idx = (self.start + self.length - 1) % len(self.data)
        result = self.data[idx]
        assert result is not None, "invariant violated"
        self.data[idx] = None
        self.length -= 1
        return result

    def appendleft(self, val: T) -> None:
        """Appends an element to the start of the deque. Raises
        BoundedDequeFullError if the deque is full
        """
        self.ensure_space_for(1)
        self.start -= 1
        if self.start == -1:
            self.start = len(self.data) - 1
        self.data[self.start] = val
        self.length += 1

    def popleft(self) -> T:
        """Pops the first element from the deque. Raises IndexError if the
        deque is empty
        """
        if self.length == 0:
            raise IndexError("pop from an empty deque")
        result = self.data[self.start]
        assert result is not None, "invariant violated"
        self.data[self.start] = None
        self.start += 1
        if self.start == len(self.data):
            self.start = 0
        self.length -= 1
        return result

    def remove(self, val: T) -> None:
        """Removes the first occurrence of val from the deque. Raises ValueError
        if val is not in the deque. This is a pretty slow operation if the
        value is in the middle
        """
        if not self:
            raise ValueError("deque is empty")

        array_idx = self.start
        real_idx = 0
        while True:
            if self.data[array_idx] == val:
                break
            real_idx += 1
            if real_idx == self.length:
                raise ValueError("value not in deque")
            array_idx += 1
            if array_idx == len(self.data):
                array_idx = 0

        if real_idx == 0:
            self.popleft()
            return
        if real_idx == self.length - 1:
            self.pop()
            return

        # need to rearrange to avoid a hole by either shifting items
        # left or right, we'll do whichever requires less movement
        hole_real_idx = real_idx
        hole_array_idx = array_idx
        del array_idx
        del real_idx

        if hole_real_idx <= (self.length // 2):
            # we have a hole at real_idx that we will fill by moving
            # every item left of the hole right 1 idx
            last = self.data[self.start]

            moving_real_idx = 1
            moving_array_idx = self.start + 1
            if moving_array_idx == len(self.data):
                moving_array_idx = 0

            while moving_real_idx <= hole_real_idx:
                tmp = self.data[moving_array_idx]
                self.data[moving_array_idx] = last
                last = tmp

                moving_real_idx += 1
                moving_array_idx += 1
                if moving_array_idx == len(self.data):
                    moving_array_idx = 0

            self.data[self.start] = None
            self.start += 1
            self.length -= 1

            if self.start == len(self.data):
                self.start = 0
            return

        # we have a hole at real_idx that we will fill by moving
        # every item right of the hole left 1 idx, replacing the right-most real index
        # with None

        moving_real_idx = hole_real_idx
        moving_array_idx = hole_array_idx

        while moving_real_idx < self.length - 1:
            next_moving_array_idx = moving_array_idx + 1
            if next_moving_array_idx == len(self.data):
                next_moving_array_idx = 0

            self.data[moving_array_idx] = self.data[next_moving_array_idx]

            moving_real_idx += 1
            moving_array_idx = next_moving_array_idx

        self.data[moving_array_idx] = None
        self.length -= 1

    def __iter__(self) -> BoundedDequeInPlaceIter[T]:
        return BoundedDequeInPlaceIter(self)



from enum import IntEnum, auto


class SubscriberToBroadcasterStatelessMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a subscriber can
    send to a broadcaster over a stateless connection. The prototypical example
    of a stateless connection would be using a distinct HTTP call for each
    message. However, this protocol just requires that there is enough structure
    to have a fixed set of variable length headers (up to 2^16 bytes per header),
    a response status code (2 bytes), and a response body (up to 2^64 bytes). If
    the response status code is not 2xx, the response body has no guarranteed format,
    otherwise, it's as indicated

    Because stateless connections generally already have more parsing helpers available,
    we just provide documentation on how the message should be structured, but don't
    actually provide parsers or serializers.

    This avoids using non-standard http headers to reduce CORS issues, instead
    embedding e.g. what would naturally be x-tracing into the response
    """

    NOTIFY = auto()
    """The subscriber is posting a message to a specific topic
    
    ### headers
    - authorization: proof the subscriber is authorized to post to the topic

    ### request body
    - 2 bytes (N): length of the topic, big-endian, unsigned
    - N bytes: the topic. if utf-8 decodable then we will attempt to match glob
      patterns, otherwise, only goes to exact subscriptions
    - 64 bytes: sha-512 hash of the message, will be rechecked
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 1 bytes (I): length of the identifier, big-endian, unsigned
    - I bytes: the identifier
    - 8 bytes (M): length of the message, big-endian, unsigned
    - M bytes: the message

    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_NOTIFY`
    """

    SUBSCRIBE_EXACT = auto()
    """The subscriber wants to receive messages posted to a specific topic. If
    the subscriber is already subscribed to the topic, the recovery url is ignored
    and no changes are made

    ### headers
    - authorization: proof the subscriber is authorized to subscribe to the topic

    ### request body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the topic, big-endian, unsigned
    - M bytes: the topic, utf-8 encoded
    - 2 bytes (R): either 0, to indicate no missed messages are desired, or the length
      of the url to post missed messages to, big-endian, unsigned
    - R bytes: the url to post missed messages to, utf-8 encoded
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data

    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """

    SUBSCRIBE_GLOB = auto()
    """The subscriber wants to receive messages to utf-8 decodable topics which match
    a given glob pattern. If the subscriber is already subscribed to the glob, the
    recovery url is ignored and no changes are made

    
    ### headers
    - authorization: proof the subscriber is authorized to subscribe to the pattern

    ### request body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the glob pattern, big-endian, unsigned
    - M bytes: the glob pattern, utf-8 encoded
    - 2 bytes (R): either 0, to indicate no missed messages are desired, or the length
      of the url to post missed messages to, big-endian, unsigned
    - R bytes: the url to post missed messages to, utf-8 encoded
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    
    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """

    UNSUBSCRIBE_EXACT = auto()
    """The subscriber wants to stop receiving messages posted to a specific topic

    ### headers
    - authorization: proof the subscriber is authorized to unsubscribe from the topic;
      formed exactly like the authorization header in SUBSCRIBE_EXACT
    
    ### request body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the topic, big-endian, unsigned
    - M bytes: the topic, utf-8 encoded
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    
    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """

    UNSUBSCRIBE_GLOB = auto()
    """The subscriber wants to stop receiving messages to utf-8 decodable topics which match
    a given glob pattern

    ### headers
    - authorization: proof the subscriber is authorized to unsubscribe from the pattern;
      formed exactly like the authorization header in SUBSCRIBE_GLOB

    ### request body
    - 2 bytes (N): length of the url that the subscriber can be reached at, big-endian, unsigned
    - N bytes: the url that the subscriber can be reached at, utf-8 encoded
    - 2 bytes (M): the length of the glob pattern, big-endian, unsigned
    - M bytes: the glob pattern, utf-8 encoded
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    
    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """

    CHECK_SUBSCRIPTIONS = auto()
    """The subscriber wants to get the strong etag representing the subscriptions for
    a specific url. Generally, the subscriber is comparing it against what it
    expects by computing it itself.

    The strong etag is the SHA512 hash of a document which is of the following
    form, where all indicated lengths are 2 bytes, big-endian encoded:
    
    ```
    URL<url_length><url>
    EXACT<topic_length><topic><recovery_length><recovery><...>
    GLOB<glob_length><glob><recovery_length><recovery><...>
    ```

    where URL, EXACT and GLOB are the ascii-representations and there
    are 3 guarranteed newlines as shown (including a trailing newline). Note
    that URL, EXACT, GLOB, and newlines may show up within
    topics/globs. The topics and globs must be sorted in (bytewise)
    lexicographical order

    ### headers
    - authorization: proof the subscriber is authorized to check the subscriptions for the url

    ### request body
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 2 bytes (N): length of the subscriber url to check, big-endian, unsigned
    - N bytes: the url to check, utf-8 encoded

    ### response body
    -  `BroadcasterToSubscriberStatelessMessageType.RESPONSE_CHECK_SUBSCRIPTIONS`
    """

    SET_SUBSCRIPTIONS = auto()
    """The subscriber wants to set all of their subscriptions and retrieve the strong etag
    it corresponds to. Unlike with `SUBSCRIBE`/`UNSUBSCRIBE`, this message is idempotent, which
    makes recovery easier. 

    The broadcaster MUST guarrantee the following properties, which mostly
    apply to concurrent requests:

    - If the subscriber is previously subscribed to a topic/glob and that
      topic/glob is in this list, the subscriber is at no point unsubscribed
      from that topic/glob due to this call
    - If the subscriber is not previously subscribed to a topic/glob and that
      is not in this list, the subscriber is at no point subscribed to that
      topic/glob due to this call
    - At some point during this call, the subscriber is subscribed to each
      (but not necessarily all) of the topics/globs in this list
    - At some point during this call, the subscriber is unsubscribed from
      each (but not necessarily all) of the topics/globs not in this list

    See the documentation for `CHECK_SUBSCRIPTIONS` for the format of the etag

    ### headers
    - authorization: proof the subscriber is authorized to set the subscriptions for the url

    ### request body
    - 2 bytes (N): length of the subscriber url to set, big-endian, unsigned
    - N bytes: the url to set, utf-8 encoded
    - 1 byte (reserved for etag format): 0
    - 64 bytes: the strong etag, will be rechecked
    - 4 bytes (E): the number of exact topics to set, big-endian, unsigned
    - REPEAT E TIMES: (in ascending lexicographic order of the topics)
      - 2 bytes (L): length of the topic, big-endian, unsigned
      - L bytes: the topic
      - 2 bytes (R): the length of the recovery url, big-endian, unsigned,
        may be 0 for no recovery url
      - R bytes: the recovery url, utf-8 encoded
    - 4 bytes (G): the number of glob patterns to set, big-endian, unsigned
    - REPEAT G TIMES: (in ascending lexicographic order of the globs)
      - 2 bytes (L): length of the glob pattern, big-endian, unsigned
      - L bytes: the glob pattern, utf-8 encoded
      - 2 bytes (R): the length of the recovery url, big-endian, unsigned,
        may be 0 for no recovery url
      - R bytes: the recovery url, utf-8 encoded

    ### response body
    - `BroadcasterToSubscriberStatelessMessageType.RESPONSE_GENERIC`
    """

    RESPONSE_CONFIRM_RECEIVE = auto()
    """The subscriber is confirming that it received a message from a broadcaster
    over a stateless connection. This primarily provides tracing data back to
    the broadcaster

    ### response body
    - 2 bytes (type): int(RESPONSE_CONFIRM_RECEIVE), big endian, unsigned
    - 2 bytes (A): length of authorization, big-endian, unsigned
    - A bytes: authorization, utf-8 encoded
    - 2 bytes (T): length of tracing data, big-endian, unsigned
    - T bytes: tracing data
    - 1 byte (I): length of the identifier, big-endian, unsigned
    - I bytes: the identifier
    - 4 bytes (N): the number of subscribers, big-endian, unsigned. usually
      1 but can be more if this subscriber forwarded the message to others
    """

    RESPONSE_CONFIRM_MISSED = auto()
    """If the broadcaster reaches out to a subscriber that it missed a message on
    a topic, the subscriber can respond in the same connection that it received
    the message. This primarily provides tracing data back to the broadcaster
    
    ### response body
    - 2 bytes (type): int(RESPONSE_CONFIRM_MISSED), big endian, unsigned
    - 2 bytes (A): length of authorization, big-endian, unsigned
    - A bytes: authorization, utf-8 encoded
    - 2 bytes (T): length of tracing data, big-endian, unsigned
    - T bytes: tracing data
    """

    RESPONSE_UNSUBSCRIBE_IMMEDIATE = auto()
    """If the broadcaster reaches out to a subscriber, the subscriber can respond in
    the same connection that it wants to unsubscribe without authorization. This
    does mean that when using a non-verifying protocol (e.g., plain http), a
    middleman can unsubscribe the subscriber from the broadcaster, but it also
    allows for recovery in a much broader set of scenarios.

    ### response body
    - 2 bytes (type): int(RESPONSE_UNSUBSCRIBE_IMMEDIATE), big endian, unsigned
    """


class BroadcasterToSubscriberStatelessMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a broadcaster can
    send to a subscriber over a stateless connection. The prototypical example
    of a stateless connection would be using a distinct HTTP call for each
    message.
    """

    RECEIVE = auto()
    """The broadcaster is notifying the subscriber of a message posted to a topic
    
    ### headers
    - authorization: proof the broadcaster can notify the subscriber
    - repr-digest: contains <digest-algorithm>=<digest>[,<digest-algorithm>=<digest>...]
      where at least one of the digest algorithms is `sha512` and the digest is the
      the base64 encoded sha-512 hash of the message

    ### request body
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 2 bytes (N): length of the topic, big-endian, unsigned
    - N bytes: the topic
    - 1 byte (I): length of the identifier, big-endian, unsigned
    - I bytes: the identifier
    - 8 bytes (M): length of the message, big-endian, unsigned
    - M bytes: the message

    ### response body
    any of:
    - `SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE`
    - `SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE`
    """

    MISSED = auto()
    """The broadcaster is notifying the subscriber they previously failed to send
    a message to the subscriber about a message on a topic the subscriber was
    subscribed to. This is an important primitive for using persistent topics
    via log channels.

    ### headers
    - authorization: proof the broadcaster can notify the subscriber

    ### request body
    - 2 bytes (T): length of the tracing data, big-endian, unsigned
    - T bytes: the tracing data
    - 2 bytes (N): length of the topic, big-endian, unsigned
    - N bytes: the topic

    ### response body
    """

    RESPONSE_GENERIC = auto()
    """
    The generic response from the broadcaster when no special data is required

    - 2 bytes (type): int(RESPONSE_GENERIC), big endian, unsigned
    - 2 bytes (A): big-endian, unsigned, the length of the authorization
    - A bytes: the authorization
    - 2 bytes (T): big-endian, unsigned, the length of tracing data
    - T bytes: the tracing data
    """

    RESPONSE_NOTIFY = auto()
    """
    The response the broadcaster sends to a subscriber after receiving a NOTIFY
    
    - 2 bytes (type): int(RESPONSE_NOTIFY), big endian, unsigned
    - 2 bytes (A): big-endian, unsigned, the length of the authorization. broadcaster
      side authorization is always used because hmac over http is supported
    - A bytes: the authorization
    - 2 bytes (T): big-endian, unsigned, the length of tracing data
    - T bytes: the tracing data
    - 2 bytes: big-endian, unsigned, the number of subscribers notified
    - 1 byte (I): length of the identifier, big-endian, unsigned
    - I bytes: the identifier
    """

    RESPONSE_CHECK_SUBSCRIPTIONS = auto()
    """The response the broadcaster sends to a subscriber after receiving a CHECK_SUBSCRIPTIONS

    - 2 bytes (type): int(RESPONSE_CHECK_SUBSCRIPTIONS), big endian, unsigned
    - 2 bytes (A): big-endian, unsigned, the length of the authorization
    - A bytes: the authorization
    - 2 bytes (T): big-endian, unsigned, the length of tracing data
    - T bytes: the tracing data
    - 1 byte (reserved for etag format): 0
    - 64 bytes: the etag
    """



"""Protocol describing something that has the semantics of asyncio.Queue"""

import asyncio
from typing import TYPE_CHECKING, Generic, Protocol, Type, TypeVar

T = TypeVar("T")


class AsyncQueueLike(Generic[T], Protocol):
    def qsize(self) -> int:
        """length of the queue; guaranteed to be accurate, as async is cooperative"""

    # maxsize purposely not allowed since asyncio.Queue uses 0 to mean unbounded,
    # but we don't want to promsie that

    def empty(self) -> bool:
        """True if the queue is empty, False otherwise"""

    def full(self) -> bool:
        """True if the queue is full, False otherwise"""

    async def put(self, item: T) -> None:
        """Put an item into the queue"""

    def put_nowait(self, item: T) -> None:
        """Put an item into the queue without blocking, raising
        asyncio.QueueFull if the queue is full
        """

    async def get(self) -> T:
        """Get an item from the queue, waiting until one is available
        if empty
        """

    def get_nowait(self) -> T:
        """Get an item from the queue without blocking, raising
        asyncio.QueueEmpty if the queue is empty
        """

    # we don't care about the task unfinished/finished stuff usually
    # since that can be accomplished is semantically simpler using events
    # and locks explicitly


if TYPE_CHECKING:
    _: Type[AsyncQueueLike] = asyncio.Queue
